# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
import pandas as pd
import pytest
import sys
import os

os.chdir("..")
ABS_PATH = os.getcwd() + os.sep + "chemeq"
sys.path.insert(0, ABS_PATH)
os.chdir(ABS_PATH)

from count_elements import count_elements
from syntax_review import syntax_review
from equation_balancer import chemeq


@pytest.mark.parametrize(
    "equation, reactants, products, elements", [
        # simple equation
        ("C2H5 + O2 = CO2 + H2O",
         ['C2H5', 'O2'],
         ['CO2', 'H2O'],
         ['C', 'H', 'O']),
        # Complex equation
        ("NH4OH + KAl(SO4)2(H2O)12 = Al(OH)3 + (NH4)2SO4 + KOH + H2O",
         ['NH4OH', 'KAl(SO4)2(H2O)12'],
         ['Al(OH)3', '(NH4)2SO4', 'KOH', 'H2O'],
         ['Al', 'H', 'K', 'N', 'O', 'S'])])
def test_syntax_review_outputs(equation, reactants, products, elements):
    assert syntax_review(equation) == (reactants, products, elements)


@pytest.mark.parametrize(
    "compound, elements, formula, prefix, count", [
        # simple compund
        ("NaCl", ["Na", "Cl"], "NaCl", 1, [1, 1]),
        # compound with subindexes
        ("H2O", ["H", "O"], "H2O", 1, [2, 1]),
        # compound with multiple subindexes
        ("C12H22O11", ["C", "H", "O"], "C12H22O11", 1, [12, 22, 11]),
        # compound with prefix
        ("35H2O", ["H", "O"], "H2O", 35, [2, 1]),
        # compound with parentheses
        ("KAl(SO4)(H2O)", ["Al", "H", "K", "O", "S"],
         "KAl(SO4)(H2O)", 1,  [1, 2, 1, 5, 1]),
        # compound with multiple subindexes, groups an prefix
        ("22KAl5(SO4)2(H2O)10", ["Al", "H", "K", "O", "S"],
         "KAl5(SO4)2(H2O)10", 22, [5, 20, 1, 18, 2])
    ])
def test_count_elements(compound, elements, formula, prefix, count):
    assert count_elements(compound, elements) == (formula, prefix, count)


def test_chemeq_parts():
    eq = chemeq("3H2 + O2 = 7H2O")
    balance = eq.is_balanced
    assert balance is False

    react_formulas = (eq.reactants['formula'].values).tolist()
    assert react_formulas == ["H2", "O2"]

    prod_formulas = (eq.products['formula'].values).tolist()
    assert prod_formulas == ["H2O"]

    counts = eq.reactants[['H','O']].values.tolist()
    assert counts == [[2, 0], [0, 2]]



# create periodic table file
import os
import sys
from importlib import util

PERIODIC_TABLE_CONTENT = """Source for the periodic table of elements:
IUPAC - International Union of Pure and Applied Chemistry
https://iupac.org/what-we-do/periodic-table-of-elements/

z,name,symbol,atomic_mass,error,group,period,state
1,Hydrogen,H,1.008,0.0002,1,1,g
2,Helium,He,4.0026,0.0001,18,1,g
3,Lithium,Li,6.94,0.06,1,2,s
4,Beryllium,Be,9.0122,0.0001,2,2,s
5,Boron,B,10.81,0.02,13,2,s
6,Carbon,C,12.011,0.002,14,2,s
7,Nitrogen,N,14.007,0.001,15,2,g
8,Oxygen,O,15.999,0.001,16,2,g
9,Fluorine,F,18.998,0.001,17,2,g
10,Neon,Ne,20.18,0.001,18,2,g
11,Sodium,Na,22.99,0.001,1,3,s
12,Magnesium,Mg,24.305,0.002,2,3,s
13,Aluminum,Al,26.982,0.001,13,3,s
14,Silicon,Si,28.085,0.001,14,3,s
15,Phosphorus,P,30.974,0.001,15,3,s
16,Sulfur,S,32.06,0.02,16,3,s
17,Chlorine,Cl,35.45,0.01,17,3,g
18,Argon,Ar,39.95,0.16,18,3,g
19,Potassium,K,39.098,0.001,1,4,s
20,Calcium,Ca,40.078,0.004,2,4,s
21,Scandium,Sc,44.956,0.001,3,4,s
22,Titanium,Ti,47.867,0.001,4,4,s
23,Vanadium,V,50.942,0.001,5,4,s
24,Chromium,Cr,51.996,0.001,6,4,s
25,Manganese,Mn,54.938,0.001,7,4,s
26,Iron,Fe,55.845,0.002,8,4,s
27,Cobalt,Co,58.933,0.001,9,4,s
28,Nickel,Ni,58.693,0.001,10,4,s
29,Copper,Cu,63.546,0.003,11,4,s
30,Zinc,Zn,65.38,0.02,12,4,s
31,Gallium,Ga,69.723,0.001,13,4,s
32,Germanium,Ge,72.63,0.008,14,4,s
33,Arsenic,As,74.922,0.001,15,4,s
34,Selenium,Se,78.971,0.008,16,4,s
35,Bromine,Br,79.904,0.003,17,4,l
36,Krypton,Kr,83.798,0.002,18,4,g
37,Rubidium,Rb,85.468,0.001,1,5,s
38,Strontium,Sr,87.62,0.01,2,5,s
39,Yttrium,Y,88.906,0.001,3,5,s
40,Zirconium,Zr,91.224,0.002,4,5,s
41,Niobium,Nb,92.906,0.001,5,5,s
42,Molybdenum,Mo,95.95,0.01,6,5,s
44,Ruthenium,Ru,101.07,0.02,8,5,s
45,Rhodium,Rh,102.91,0.01,9,5,s
46,Palladium,Pd,106.42,0.01,10,5,s
47,Silver,Ag,107.87,0.01,11,5,s
48,Cadmium,Cd,112.41,0.01,12,5,s
49,Indium,In,114.82,0.01,13,5,s
50,Tin,Sn,118.71,0.01,14,5,s
51,Antimony,Sb,121.76,0.01,15,5,s
52,Tellurium,Te,127.6,0.03,16,5,s
53,Iodine,I,126.9,0.01,17,5,s
54,Xenon,Xe,131.29,0.01,18,5,g
55,Cesium,Cs,132.91,0.01,1,6,s
56,Barium,Ba,137.33,0.01,2,6,s
72,Hafnium,Hf,178.49,0.01,4,6,s
73,Tantalum,Ta,180.95,0.01,5,6,s
74,Tungsten,W,183.84,0.01,6,6,s
75,Rhenium,Re,186.21,0.01,7,6,s
76,Osmium,Os,190.23,0.03,8,6,s
77,Iridium,Ir,192.22,0.01,9,6,s
78,Platinum,Pt,195.08,0.02,10,6,s
79,Gold,Au,196.97,0.01,11,6,s
80,Mercury,Hg,200.59,0.01,12,6,l
81,Thallium,Tl,204.38,0.01,13,6,s
82,Lead,Pb,207.2,1.1,14,6,s
83,Bismuth,Bi,208.98,0.01,15,6,s"""


CURRENT_PATH = os.path.abspath(".")
installed = util.find_spec("chemeq")

# Different imports if package is installed or in source directory
if installed.has_location:
    # if installed run installed version
    INSTALLATION_PATH = installed.submodule_search_locations[0] + os.sep
    os.chdir(INSTALLATION_PATH)
    if "periodic_table.csv" not in os.listdir():
        with open('periodic_table.csv', 'w') as pt:
            pt.write(PERIODIC_TABLE_CONTENT)

    # shorten import paths
    from chemeq.equation_balancer import chemeq
    from chemeq.equation_balancer import periodic_table
    from chemeq.syntax_review import syntax_review
    from chemeq.count_elements import count_elements
    os.chdir(CURRENT_PATH)
else:
    # if not installed run from source location
    sys.path.insert(0, CURRENT_PATH)
    from equation_balancer import chemeq
    from equation_balancer import periodic_table
    from syntax_review import syntax_review
    from count_elements import count_elements


__all__ = [
            "chemeq",
            "periodic_table",
            "count_elements",
            "syntax_review"
          ]
__author__ = "Elbio Peña Almonte"



import os
import re
import pandas as pd
from importlib import util

CURRENT_PATH = os.path.abspath(".")

# if package is installed change to installation directory
installed = util.find_spec("chemeq")
if installed.has_location:
    INSTALLATION_PATH = installed.submodule_search_locations[0] + os.sep
    os.chdir(INSTALLATION_PATH)

periodic_table = pd.read_csv("periodic_table.csv", header=3)
os.chdir(CURRENT_PATH)


def syntax_review(equation):
    '''Verify the chemical equation syntax and if it is correct, returns
    the reactants and products compounds and a list of the present elements

    equation: type str. represents chemical equation.
    all_symbols: type list[str], symbols for all chemical elements.
    return: type list[str], reactants formulas witout prefix (molar values),
            type list[str]. Products formulas witout prefix (molar values),
            type list[str]. Elements present in the chemical equation
    :rtype: list[ str], list[str], list[str]

    :raises TypeError: if :param:'equation' is not str
    :raises ValueError: if symbols other than chemical elelments are used
                        eg. na is not a valid element but Na is.
    :raises ValueError: if there is not an equal sign in equation.
    :raises ValueError: if there are more than one equal sign.
    :raises ValueError: if there are open parentheses within a compound.
    :raises ValueError: if there are empty parentheses.
    :raises ValueError: if two plus(+) signs are contiguous.
    :raises ValueError: if one side of the equation is empty.
    :raises ValueError: if there are diferent elements in the sides of equation
    '''
    all_symbols = periodic_table['symbol'].values

    def get_elements(compound, all_symbols):
        '''returns a list with all elements present in a compound

        :param compound: chemical compound
        :param type:  str

        :returns elements: lis of elements present in the compound
        :rtype: list[str]

        examples
        --------
        .. code block: python
            >>> get_elements("H2O", all_symbols)
            ['H','O']
            >>> get_elements("Cu(CO3)(OH)2", all_symbols)
            ['C', 'Cu', 'H', 'O']'''

        elements_in_comp = []

        # separate  subgroups of molecule and find their elements
        pattern = re.compile(r"(\()([A-Z]+[a-z]*[0-9]*)+(\))[0-9]*")
        subgroups = [x.group() for x in pattern.finditer(compound)]
        for subgroup in subgroups:
            compound = compound.replace(subgroup, "")
        if len(compound):
            subgroups = [compound] + subgroups
        subgroups = [x.replace("(", "") for x in subgroups]
        for subgroup in subgroups:
            # find words beginning in uppercase followed by lowercase/numbers
            # like O, Be, H20 and check  if they match an element.
            pattern = re.compile("[A-Z]{1}[a-z]*[0-9]*|[a-z]+[0-9]*")
            matches = pattern.finditer(subgroup)
            for found in matches:
                symbol = re.match(r"^[A-Z]*[a-z]*|\W*", found.group()).group()
                if symbol in all_symbols:
                    elements_in_comp += [symbol]
                else:
                    text = f'Unrecognized element: "{symbol}" in "{compound}"'
                    raise ValueError(text)
        return list(elements_in_comp)

    def unique(list_of_elements):
        singles = []
        for element in list_of_elements:
            if element not in singles:
                singles += [element]
        return singles

    ########################################################################
    ########################################################################

    # Check equation is a string
    if not isinstance(equation, str):
        text = "Equation must be a string not a "
        raise TypeError(text + f"{type(equation)}")

    # look for characters other than valid characters.
    wrong_symbols = list(re.findall(r"[^a-z0-9A-Z ()\+=]+",
                                    equation))
    if len(wrong_symbols):
        text = 'Unrecognized symbol: '
        for symbol in wrong_symbols:
            text += '"' + symbol + '", '
        raise ValueError(text[:-2])

    if len(list(re.findall("(=)", equation))) > 1:
        raise ValueError("Too many equal signs")

    try:
        left_hand_side, right_hand_side = equation.replace(
                                            " ", "").split("=")
    except ValueError:
        raise ValueError('Missing equal sign "=" in equation')

    if len(left_hand_side) == 0:
        raise ValueError(f'Reactants missing in:  "{equation}"')

    if len(right_hand_side) == 0:
        raise ValueError(f'Products missing in: "{equation}"')

    reac_formulas = left_hand_side.split("+")
    prod_formulas = right_hand_side.split("+")

    # if there are empty compounds '', there are two contiguous (+) signs
    if ('' in reac_formulas) | ('' in prod_formulas):
        text = 'there are too many plus (+) signs'
        raise ValueError(text)

    for comp in (reac_formulas + prod_formulas):
        if len(re.findall(r"(\(\))+", comp)):
            raise ValueError(f'Empty parentheses in "{comp}"')

        # verify every parenthesis is closed.
        parentheses = list(re.findall("[(]{1}|[)]{1}", comp))
        if len(parentheses):
            state = 0
            for item in parentheses:
                if item == '(':
                    state += 1
                else:
                    state += -1
            if state != 0:
                text = 'Open parentheses in: "' + comp + '"'
                raise ValueError(text)

    # find elements in both sides of eq and verify they are the same type
    reac_elements = []
    for comp in reac_formulas:
        for el in unique(get_elements(comp, all_symbols)):
            if el not in reac_elements:
                reac_elements += [el]
    reac_elements.sort()
    prod_elements = []
    for comp in prod_formulas:
        for el in unique(get_elements(comp, all_symbols)):
            if el not in prod_elements:
                prod_elements += [el]
    prod_elements.sort()
    if reac_elements != prod_elements:
        text = "Reactant's elements are different to product's elements"
        raise ValueError(text)
    return reac_formulas, prod_formulas, reac_elements



import setuptools


LONG_DESC = open("README.md", "r").read()

setuptools.setup(
    name='chemeq',
    version='0.1.7',
    description="Balance chemical equations & calculates molecular weights plus periodic table of elements",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    author="Elbio Peña",
    author_email="elbioemilio@outlook.es",

    packages=setuptools.find_packages(where="."),
    package_dir={"": "."},
    package_data={"chemeq": ["chemeq/*.csv"], "": ["README.md"]},
    include_package_data=True,
    exclude_package_data={"chemeq": [".gitignore"]},
    python_requires=">= 3.6",
    install_requires=["numpy", "pandas"],
    keywords=["chemistry", "chemical", "chemicals", "equation",
              "equations", "balance", "balancer", "balancing", "reaction",
              "reactions", "periodic table", "elements"],

    classifiers=["Development Status :: 5 - Production/Stable",
                 "Programming Language :: Python :: 3",
                 "Operating System :: OS Independent",
                 "Intended Audience :: Education",
                 "Intended Audience :: End Users/Desktop"],
    url='https://github.com/elbiop/chemeq'
)



# -*- coding: utf-8 -*-
import pytest
import sys
import os
os.chdir("..")
ABS_PATH = os.getcwd() + os.sep + "chemeq"
sys.path.insert(0, ABS_PATH)
os.chdir(ABS_PATH)
from syntax_review import syntax_review


def test_syntax_review():
    # Must be a string
    with pytest.raises(TypeError):
        syntax_review(1234)

    with pytest.raises(ValueError):
        # Only valid characters: +, =, Letters, numbers, perentheses.
        syntax_review("#$% + *&^ = #$%*&")
        # must have at least one equal sign
        syntax_review("H2+O2 H2O")
        # only one equal sign
        syntax_review("H2+O2=H2O=H2O")
        # Must have reactants
        syntax_review("=H2O")
        # Must have products
        syntax_review("H2+O2=")
        # canot have empty parentheses
        syntax_review("C6H9()(OOH)3+O2 =CO2+H2O")
        # cannot have open pearentheses
        syntax_review(")H2+O2=H2O")
        # cannot have open pearentheses
        syntax_review("(H2+O2=)H2O")
        # case sensitive
        syntax_review("Ca+O2=caO")
        # reactants and products have the same elements
        syntax_review("H2+O2=NaCl")
        # Only periodic table symbols
        syntax_review("abc = abc")
        # Reactants and Product elements must be the same
        syntax_review("Na + Cl = KCl")



import re


def count_elements(compound, elements):
    '''counts the number of appearances of a list of elements in a compound

     param compound: type str. Chemical compound.
     param elements: list[str], symbols for all chemical elements
     return: compound without witout prefix (molar value),
              prefix (molar value) or 1 if not present,
              elements count per one molecule.
     rtype: str, int, list[int]

     examples
     --------
         >>> count_elements("H2O",['H','O'])
         ('H2O', 1, [2, 1])
         >>> count_elements("3Cu(CO3)(OH)2",['C', 'Cu', 'H', 'O'])
         ('Cu(CO3)(OH)2', 3, [1, 1, 2, 5])
     '''
    original_comp = compound
    subindexes = [0 for _ in range(len(elements))]
    # find if a compound begins with a number
    coeff = re.findall(r"^[0-9]+(?=[A-Z]|\()", compound)
    if len(coeff):
        trim = len(coeff[0])
        compound = compound[trim:]
        coeff = int(coeff[0])
    else:
        trim = 0
        coeff = 1

    # find groups that are between parentheses and separate them into subgroups
    # iterate over each subgroup of the compound
    pattern = re.compile(r"(\()([A-Z]+[a-z]*[0-9]*)+(\))[0-9]*")
    subgroups = [x.group() for x in pattern.finditer(compound)]
    for subgroup in subgroups:
        compound = compound.replace(subgroup, "")
    if len(compound):
        subgroups = [compound] + subgroups
    subgroups = [x.replace("(", "") for x in subgroups]
    for subgroup in subgroups:
        group_sub_ix = 1
        if subgroup.find(")") > 0:
            subgroup, remain = subgroup.split(")")
            if len(remain):
                group_sub_ix = int(remain)
        '''find the element that is followed by 0 or more numbers and
        without retrieval not followed by lowercase to avoid false positive
        of single letters elements with two letter elements that begin with
        the same letter'''
        # add whitespace to comply with the regex in case of no sub index
        subgroup += " "
        for k, el in enumerate(elements):
            # Find uppercases followed or not by a number and (do not include)
            # not followed by lowercase letters,
            pattern = re.compile(f'{el}' + '{1}[0-9]*(?=[^a-z])')
            matches = list(pattern.finditer(subgroup))
            for m in matches:
                # if match is not followed by numbers add 1
                # if is add that number"
                if len(m.group()[len(el):]):
                    subindexes[k] += group_sub_ix*int(m.group()[len(el):])
                else:
                    subindexes[k] += group_sub_ix
    return(original_comp[trim:], coeff, subindexes)



import os
import sys
import numpy as np
import pandas as pd
from importlib import util

CURRENT_PATH = os.path.abspath(".")
installed = util.find_spec("chemeq")

# Different imports if package is installed or in source directory
if installed.has_location:
    # if installed run installed version
    INSTALLATION_PATH = installed.submodule_search_locations[0] + os.sep
    os.chdir(INSTALLATION_PATH)
else:
    # if not installed run from source location
    sys.path.insert(0, CURRENT_PATH)

from syntax_review import syntax_review
from count_elements import count_elements
os.chdir(CURRENT_PATH)

periodic_table = pd.read_csv("periodic_table.csv", header=3)


class chemeq():
    '''CLASS that receives a string representing a chemical equation of the
    form:
    "reactant_1 + ... + reactant_n  = product_1 + ... + product_n"

    LIMITATIONS
    1. Does not include Rare earths, synthetic elements, Rn, Fr & Ra.
    2. Intermediate prefixes like those from hydrates such as Cu(SO4)·5H2O
          must be represented with subindexes like Cu(SO4)(H2O)5.

    EXAMPLES
    if an unbalanced equation is provided:
    >>> eq = chemeq("C2H5(OH) + O2 = CO2 + H2O")
    >>> eq
    <Unbalanced equation: "C2H5(OH) + O2 = CO2 + H2O">

    If a balanced equation is provided:
    >>> eq = chemeq("3O2 + C2H5(OH) = 2CO2 + 3H20")
    >>> eq
    <Balanced equation: "3O2 + C2H5(OH) = 2CO2 +3H2O">'''

    def __init__(self, equation):
        self.__eq = equation.replace(" ", "")
        self.__react = []
        self.__react = pd.DataFrame()
        self.__prod = pd.DataFrame()
        self.__balanced = False
        self.__extract()
        self.__reformat()
        self.__checkbal(self.__react['coefficient'].values,
                        self.__react[self.__elem__].values,
                        self.__prod['coefficient'].values,
                        self.__prod[self.__elem__].values)

    def __str__(self):
        return self.__eq

    def __repr__(self):
        if self.__balanced:
            return f'<Balanced equation: "{self.__eq}">'
        else:
            return f'<Unbalanced equation: "{self.__eq}">'

    @property
    def is_balanced(self):
        'Values are boolean: True or False'
        return self.__balanced

    @property
    def reactants(self) -> pd.DataFrame:
        '''pandas.DtataFrame where each row represents one of the reactant
        compounds, its molecular weights and elements.

        EXAMPLE:
        >>>eq = chemeq("C2H5(OH) + 3O2 = 2CO2 + 3H2O")
        >>> eq.reactants

                coefficient   formula  C  H  O  Mass(g/mol)
              0            1  C2H5(OH)  2  6  1       46.069
              1            3        O2  0  0  2       31.998'''
        return self.__react

    @property
    def products(self) -> pd.DataFrame:
        '''pandas.DtataFrame where each row represents one of the product
        compounds, molecular weights and elements.

        EXAMPLE:
        >>>eq = chemeq("C2H5(OH) + 3O2 = 2CO2 + 3H2O")
        >>>eq.products

            coefficient formula  C  H  O  Mass(g/mol)
          0            2     CO2  1  0  2       44.009
          1            3     H2O  0  2  1       18.015'''
        return self.__prod

    def __reformat(self):
        '''reassemble the equation with single spaces between  and
        symbols '''
        eq = ""
        for i, comp in enumerate(self.__react['formula']):
            prefix = int(self.__react['coefficient'][i])
            if prefix == 1:
                prefix = ""
            else:
                prefix = str(prefix)
            eq += prefix + comp + " + "
        eq = eq[:-2]
        eq += "= "
        for i, comp in enumerate(self.__prod['formula']):
            prefix = int(self.__prod['coefficient'][i])
            if prefix == 1:
                prefix = ""
            else:
                prefix = str(prefix)
            eq += prefix + comp + " + "
        self.__eq = eq[:-3]

    def __extract(self):
        '''Extract relevant information from equation such as reactant
        compounds, product compounds, Elements in each, the total quantity of
        each element and the respetive molecular masses of reactants and
        compounds'''
        all_symbols = periodic_table['symbol']
        reac, prod, elem = syntax_review(self.__eq)
        self.__react['formula'] = reac
        self.__prod['formula'] = prod
        self.__elem__ = elem
        atomic_masses = periodic_table[all_symbols.isin(
                                    self.__elem__)].sort_values(by="symbol")
        atomic_masses = atomic_masses['atomic_mass'].values

        # Extract coefficients formula and elements count
        formulas, coefficients, counts = [], [], []
        for i, comp in enumerate(self.__react['formula']):
            formula, coefficient, count = count_elements(comp, self.__elem__)
            formulas += [formula]
            coefficients += [coefficient]
            counts += [count]
        self.__react = pd.DataFrame(np.array(coefficients)[:, np.newaxis],
                                    columns=['coefficient'])
        self.__react['formula'] = formulas
        self.__react = pd.concat([self.__react,
                                  pd.DataFrame(counts,
                                               columns=self.__elem__,
                                               dtype=np.int64)], axis=1)
        self.__react['Mass(g/mol)'] = np.sum(np.multiply(counts,
                                             atomic_masses), axis=1)

        formulas, coefficients, counts = [], [], []
        for i, comp in enumerate(self.__prod['formula']):
            formula, coefficient, count = count_elements(comp, self.__elem__)
            formulas += [formula]
            coefficients += [coefficient]
            counts += [count]

        self.__prod = pd.DataFrame(np.array(coefficients)[:, np.newaxis],
                                   columns=['coefficient'])
        self.__prod['formula'] = formulas
        self.__prod = pd.concat([self.__prod,
                                 pd.DataFrame(counts,
                                              columns=self.__elem__,
                                              dtype=np.int64)],
                                axis=1)
        self.__prod['Mass(g/mol)'] = np.sum(np.multiply(counts,
                                            atomic_masses), axis=1)

    def __checkbal(self, reac_coef, reac_count, prod_coef, prod_count):
        '''check if elements in both sides of eq are in equal amounts.'''
        reac_atoms = np.matmul(reac_coef, reac_count)
        prod_atoms = np.matmul(prod_coef, prod_count)
        if np.multiply.reduce(reac_atoms == prod_atoms):
            self.__balanced = True
        else:
            self.__balanced = False
        return (reac_atoms, prod_atoms)

    def balance(self):
        '''balances a chemical equation of the form:
        "compound_1 + ... + compound_n  = product_1 + ... + product_n"


        limitations
        -----------
        - Does not include Rare earths, synthetetic elements, Rn, Fr & Ra.
        - Hidrates representations must have coefficient like Cu(SO4)(H2O)5
          instead of prefixes like Cu(SO4)·5H2O

        examples
        --------
        >>> eq = chemeq('C2H6 + O2 = CO2 + H2O')
        >>> eq
            <Unbalanced equation: "C2H6 + O2 = CO2 + H2O">
        >>> eq.balance()
        >>> eq
            <Balanced equation: '2C2H6 + 7O2 = 4CO2 + 6H2O'>
        '''
        # separate the compounds of each side of equation
        # create the matrices of subindexes and coefficients of both sides
        # each row will be a compound and each column one of the elements
        #         that are present in the equation

        if self.__balanced:
            return None

        reac_count = self.__react[self.__elem__].values.astype(np.int64)
        prod_count = self.__prod[self.__elem__].values.astype(np.int64)
        reac_coef = np.full(reac_count.shape[0], 1, dtype=np.int64)
        prod_coef = np.full(prod_count.shape[0], 1, dtype=np.int64)

        ''' ****Algorithm****
        if the equation is unbalanced the sum of elements in one side will be
        not equal to the elements in the other side
            choose the side with more elements ( reac or prod )
            find which element has a different sum
            pick one at random.
            find the compounds in the other side that have that element
            Update the coefficient of that compund to counteract the difference
            check if it is balanced if not repeat
            '''
        # the maximum coefficient in the intermediate equation while is being
        # balanced can be higher than the final coefficient but is proportional
        # to the maximum subindex in the equation
        max_coef = 50*np.sqrt(np.max(np.hstack([np.max(reac_count, axis=0),
                                                np.max(prod_count, axis=0)])))
        # Maximum number of iterations
        high_limit = 50_000
        counter = 0
        reac_atoms, prod_atoms = self.__checkbal(reac_coef, reac_count,
                                                 prod_coef, prod_count)

        while not(self.__balanced):
            # if right side has more elements count
            #     find the coefficient and the coefficient of the element that
            #     has more elements and multiply by the compound in the left
            #     side with that element update the coefficients

            if (True in (prod_atoms > reac_atoms)) and (True in (reac_atoms >
                                                                 prod_atoms)):
                out_of_balance = ['reactants',
                                  'products'][np.random.randint(2)]
            elif True in (prod_atoms > reac_atoms):
                out_of_balance = 'products'
            else:
                out_of_balance = 'reactants'

            if out_of_balance == 'products':
                # What elements are more abundant in the product?
                elem_index = np.where(prod_atoms > reac_atoms)[0]
                # pick at random one of the elements that are in excess
                elem_index = elem_index[np.random.randint(len(elem_index))]
                # pick at random a compound of reactants that has that atom
                # compounds where the sub is > 0, i.e. the atom is present
                comp_index = np.where(reac_count[:, elem_index] > 0)[0]
                comp_index = comp_index[np.random.randint(len(comp_index))]
                # how many elements has to be added to the coefficient to equal
                #    the elements in the other side take the difference between
                #    the number of elements in both sides and divide by the
                #    sub-index of the element in the selected compound
                increment = (prod_atoms - reac_atoms)[elem_index]/reac_count[
                                                        comp_index, elem_index]
                try:
                    if increment % 1 > 0:
                        reac_coef = reac_coef + reac_count[comp_index,
                                                           elem_index]

                        reac_coef[comp_index] = (reac_coef[comp_index] +
                                                 (prod_atoms - reac_atoms)
                                                 [elem_index] - reac_count
                                                 [comp_index, elem_index]
                                                 )
                    else:

                        reac_coef[comp_index] = (reac_coef[comp_index] +
                                                 increment
                                                 )
                except Exception:
                    # In case the increment overflows int64 capacity
                    pass

            elif out_of_balance == 'reactants':
                # What elements are more abundant in the reactants?
                elem_index = np.where(reac_atoms > prod_atoms)[0]
                # pick a random element that is in excess
                elem_index = elem_index[np.random.randint(len(elem_index))]
                # pick a random compound of the products that has that element
                # compounds where the sub is > 0, i.e. the element is present
                comp_index = np.where(prod_count[:, elem_index] > 0)[0]
                comp_index = comp_index[np.random.randint(len(comp_index))]
                # how many elements has to be added to the coefficient to equal
                #    the elements in the other side take the difference between
                #    the number of elements in both sides and divide by the
                #    sub-index of the element in the selected compound
                increment = (reac_atoms - prod_atoms)[elem_index]/prod_count[
                                                    comp_index, elem_index]
                try:
                    if increment % 1 > 0:
                        prod_coef = prod_coef + prod_count[comp_index,
                                                           elem_index]
                        prod_coef[comp_index] = (prod_coef[comp_index] + (
                                    reac_atoms - prod_atoms)[elem_index] -
                                    prod_count[comp_index, elem_index])
                    else:
                        prod_coef[comp_index] = (prod_coef[comp_index] +
                                                 increment)
                except Exception:
                    # In case the increment overflows int64 capacity
                    pass
            reac_atoms, prod_atoms = self.__checkbal(
                            reac_coef, reac_count, prod_coef, prod_count)
            # break here before restarting coefficients in case is balanced
            if self.__balanced:
                break

            # force stop if reach high limit
            counter += 1
            if counter == high_limit:
                text = f"The equation couldn't be solved, {high_limit:,}"
                raise RuntimeError(text + " iterations reached!")

            # if the coeficents reach a number to high restart the coefficients
            if np.max(np.hstack([reac_coef, prod_coef])) > max_coef:
                reac_coef = np.full(len(reac_coef), 1)
                prod_coef = np.full(len(prod_coef), 1)

        # Simplify coeficients dividing by the greatest common divider
        reac_coef = np.array(reac_coef, np.int64)
        prod_coef = np.array(prod_coef, np.int64)
        gcd = np.gcd.reduce(np.hstack([reac_coef, prod_coef]))
        reac_coef = reac_coef/gcd
        prod_coef = prod_coef/gcd
        self.__react['coefficient'] = reac_coef.astype(np.int64)
        self.__prod['coefficient'] = prod_coef.astype(np.int64)
        self.__reformat()
